#include <stdio.h>
#include <qapplication.h>
#include <qlistbox.h>
#include <qlineedit.h>
#include <qfiledialog.h>
#include <qlabel.h>
#include <qdom.h>
#include <qtextbrowser.h>
#include <qclipboard.h>
#include <qsocketnotifier.h>
#include "moc_pingvo.h"
#include "X11/Xlib.h"

extern int recorder_datafd;
void recorder_init();
void recorder_cleanup();
void do_recording();
void recorder_process_replies();

IdxViewItem::IdxViewItem( int start, int end, char *text0 ):
	art_start(start), art_end(end), tstr(text0) {
}

QString IdxViewItem::text() const {
  return tstr;
}

IdxView::IdxView( QWidget *parent, const char *name, WFlags f):
	IdxViewUI( parent, name, f ),
	dname(QString::null), idxname(QString::null),
	rootNode(0), xmlMode(false), configChanged(false),
	defArtX(0), defArtY(0), defArtWidth(420), defArtHeight(470),
	artBrowsersNum(0), configSaveTimer(0), configured(false) {
  setFixedSize( size() );
  loadConfig();
  show();
  qApp->processEvents();
  loadConfig();
  recorder_init();
  ev_sn = new QSocketNotifier( recorder_datafd, QSocketNotifier::Read, this );
  connect( ev_sn, SIGNAL(activated(int)), SLOT(xRecordEventsAvail()) );
  do_recording();
  configured = true;
}

IdxView::~IdxView() {
  delete ev_sn;
  recorder_cleanup();
//  if ( configChanged ) saveConfig();
}

IdxViewItem *IdxView::lookupNode( IdxViewItem *from,
	QString prefix, int pfxlen ) {
  IdxViewItem *n, *pnode;

  n = from;
  pnode = 0;
  while ( n ) {
  	if ( n->tstr.left( pfxlen ).lower() == prefix.left( pfxlen ) ) {
		if ( pfxlen < prefix.length() && n->firstChild ) {
			pfxlen++;
			pnode = n;
			n = n->firstChild;
		}
		else return n;
	}
	else n = n->nextSibling;
  }

  return pnode;
}

IdxViewItem *IdxView::lookupNode( QString prefix ) {
  if ( !rootNode ) return 0;
  return lookupNode( rootNode, prefix.lower(), 1 );
}

void IdxView::insertNode( IdxViewItem *node ) {
  IdxViewItem *n;

  node->firstChild = 0;
  n = lookupNode( node->tstr );
  if ( !n ) {
	if ( !rootNode ) {
		rootNode = node;
		node->prevSibling = 0;
		node->nextSibling = 0;
		return;
	}
	node->prevSibling = 0;
	node->nextSibling = rootNode;
	rootNode->prevSibling = node;
	rootNode = node;
  }

  if ( node->tstr.length() < n->tstr.length() ) {
  	if ( rootNode == n ) rootNode = node;
	node->prevSibling = n->prevSibling;
	if ( node->prevSibling )
		node->prevSibling->nextSibling = node;
	node->nextSibling = n->nextSibling;
	node->firstChild = n;
	n->prevSibling = node;
	n->nextSibling = 0;
  }
  else {
  	node->prevSibling = n;
  	node->nextSibling = n->firstChild;
	n->firstChild = node;
  }
  dumpTree();
}

void IdxView::dumpChilds( IdxViewItem *from, int indent ) {
  int i;
  IdxViewItem* node = from->firstChild;

  while ( node ) {
	for ( i = 0; i < indent; i++ ) printf( " " );
	printf( "%s\n", node->tstr.latin1() );
	dumpChilds( node, indent+1 );
	node = node->nextSibling;
  }
}

void IdxView::dumpTree() {
  if ( !rootNode ) {
  	puts( "(empty tree)" );
  	return;
  }

  puts( "----" );
  printf( "%s\n", rootNode->tstr.latin1() );
  dumpChilds( rootNode, 1 );
  puts( "----" );
}

void IdxView::loadDict( QString prefix ) {
  FILE *f;
  int pos, rc, pfxlen;
  char *buf;
  IdxViewItem *node, *prev;

  if ( prefix ) pfxlen = strlen( prefix );
  f = fopen( dname.latin1(), "r" );
  if ( !f ) {
  	perror( "fopen" );
  	return;
  }

  prev = 0;
  while ( !feof(f) ) {
	rc = fscanf( f, "%d %a[^\n]", &pos, &buf );
	if ( rc != 2 ) break;
	if ( prefix && *prefix && strncasecmp( buf, prefix, pfxlen ) ) {
		free( buf );
		continue;
	}
	node = new IdxViewItem( pos, -1, buf );
	lb->insertItem( node );
//	insertNode( node );
	free( buf );
	if ( prev ) prev->art_end = pos;
	prev = node;
  }
  fclose( f );
//  dumpTree();
}

void IdxView::loadDictXML( QString prefix ) {
  IdxViewItem *node;
  QFile f( wfname );
  int idx = 0;

#if 1
  char buf[80];
  int len;

  if ( !f.open( IO_ReadOnly ) ) {
  	puts( "pingvo: failed to open wordlist" );
	return;
  }
  
  while ( !f.atEnd() ) {
    len = f.readLine( buf, sizeof buf );
    if ( len == -1 ) {
    	puts( "pingvo.cc wordlist reading failure" );
	return;
    }

    if ( buf[len-1] == '\n' ) buf[len-1] = 0;
    node = new IdxViewItem( -1, -1, buf );
    node->idx = idx++;
    lb->insertItem( node );
  }
  f.close();

#else
  QDomDocument doc("dictionary");
  QDomElement docEl, e;
  QDomNode n;
  QString pfx = prefix.lower();

  if ( !f.open( IO_ReadOnly ) ) {
  	puts( "pingvo: failed to open dictionary" );
	return;
  }
  doc.setContent( &f );
  f.close();
  docEl = doc.documentElement();
  for ( n = docEl.firstChild(); !n.isNull(); n = n.nextSibling() ) {
	if ( !n.isElement() ) continue;
	e = n.toElement();
	if ( e.tagName() != "entry" ||
	     e.attribute("word").left(pfx.length()).lower() != pfx ) continue;
	node = new IdxViewItem( -1, -1, (char*) e.attribute("word").latin1() );
	node->idx = idx++;
	lb->insertItem( node );
  }
#endif
  if ( lb->firstItem() ) {
  	lb->setSelected( lb->firstItem(), TRUE );
  	lb->setCurrentItem( lb->firstItem() );
	le->clear();
  }
}

void IdxView::lookupByPrefix() {
#if 1
  QListBoxItem *it;
  QString prefix;
  int pfxlen;
  
  prefix = le->text();
  pfxlen = prefix.length();
  it = lb->firstItem();
  while ( it ) {
  	if ( it->text().left(pfxlen) == prefix ) {
		lb->setSelected( it, true );
		lb->setCurrentItem( it );
		lb->ensureCurrentVisible();
		break;
	}
	it = it->next();
  }
#else
  lb->clear();
  if ( xmlMode )
  	loadDictXML( /*le->text()*/ "" );
  else
	loadDict( le->text() );
  lb->setSelected( lb->firstItem(), true );
#endif
}

void IdxView::loadEnRu() {
  lb->clear();
  dname = "lingvo-universal-en-ru.dict";
  idxname = "lingvo-universal-en-ru.idx";
  loadDict( le->text() );
  lb->setSelected( lb->firstItem(), true );
  xmlMode = false;
}

void IdxView::loadRuEn() {
  lb->clear();
  dname = "";
  idxname = "";
  xmlMode = false;
}

void IdxView::loadDefXML() {
  lb->clear();
  dname = "lingvo.xml";
  idxname = "lingvo.idx";
  wfname = "lingvo.words";
  loadDictXML( /*le->text()*/ "" );
  xmlMode = true;
}

void IdxView::loadCustom() {
  QString s = QFileDialog::getOpenFileName();
  if ( s.isEmpty() ) return;
  lb->clear();
  dname = s;
  if ( dname.right(5) == ".dict" ) {
  	idxname = dname.left( dname.length() - 5 ) + ".idx";
	loadDict( le->text() );
	xmlMode = false;
  }
  else if ( dname.right(4).lower() == ".xml" ) {
  	idxname = dname.left( dname.length() - 4 ) + ".idx";
	wfname = dname.left( dname.length() - 4 ) + ".words";
  	loadDictXML( /*le->text()*/ "" );
	xmlMode = true;
  }
  lb->setSelected( lb->firstItem(), true );
}

QDomElement IdxView::getArticleXML( IdxViewItem *it ) {
  QDomDocument doc("dictionary");
  QDomElement docEl, e;
  QDomNode n;
  QFile f( idxname ), f2( dname );
  char *buf;
  int idxrec[2];

  if ( !f.open( IO_ReadOnly ) ) {
  	puts( "pingvo: failed to open dictionary index" );
	return QDomElement();
  }
  f.at( it->idx * sizeof idxrec );
  f.readBlock( (char*) idxrec, sizeof idxrec );
  f.close();
  
  if ( !f2.open( IO_ReadOnly ) ) {
  	puts( "pingvo.cc failed to open dictionary" );
	return QDomElement();
  }
  f2.at( idxrec[0] );
  buf = new char[idxrec[1]];
  f2.readBlock( buf, idxrec[1] );
  f2.close();

  doc.setContent( QCString(buf) );
  return doc.documentElement();
}

void IdxView::showCurrentArticle() {
  QListBoxItem *it = lb->selectedItem();
  if ( !it ) return;
  showArticle( it );
}

void IdxView::showArticle( QListBoxItem *lbItem ) {
  ArticleView *av;
  IdxViewItem *node;

  node = (IdxViewItem*) lbItem;
  if ( xmlMode ) {
  	QDomElement e = getArticleXML( node );
	if ( e.isNull() ) return;
  	av = new ArticleView();
	connect( av, SIGNAL(geometryOnClose(int,int,int,int)), SLOT(articleClosed(int,int,int,int)) );
	connect( av, SIGNAL(anchorClicked(const QString&, const QString&)),
		SLOT(processAnchorClick(const QString&, const QString&)) );

	if ( !artBrowsersNum ) av->move( defArtX, defArtY );
	av->resize( defArtWidth, defArtHeight );
	
	av->loadArticleXML( e );
	qApp->processEvents();

	if ( !artBrowsersNum ) av->move( defArtX, defArtY );
	av->resize( defArtWidth, defArtHeight );

	artBrowsersNum++;
  }
}

void IdxView::keyPressEvent( QKeyEvent *e ) {
  QListBoxItem *it;

  if ( e->key() == Key_Up || e->key() == Key_Down ||
       e->key() == Key_PageUp || e->key() == Key_PageDown ) {
  	qApp->sendEvent( lb, e );
	qApp->processEvents();
	le->setText( lb->currentText() );
	le->selectAll();
  }
}

void IdxView::moveEvent( QMoveEvent *e ) {
  if ( !configured ) return;
  configChanged = true;
  if ( !configSaveTimer ) {
  	configSaveTimer = new QTimer( this );
	connect( configSaveTimer, SIGNAL(timeout()), SLOT(confSaveTimeout()) );
	configSaveTimer->start( 500, TRUE );
  }
}

void IdxView::closeEvent( QCloseEvent *e ) {
  configChanged = true;
  saveConfig();
  e->accept();
}

void IdxView::loadConfig() {
  QDomDocument doc;
  QDomElement docEl, e;
  QDomNode n;
  QFile f( "config.xml" );
  bool idxpos_initialized = false, artgeom_initialized = false;

  if ( !f.open( IO_ReadOnly ) ) return;
  doc.setContent( &f );
  n = doc.documentElement().firstChild();
  while ( !n.isNull() ) {
  	e = n.toElement();
	if ( e.tagName() == "IdxWinPos" && !idxpos_initialized ) {
		move( e.attribute("x").toInt(), e.attribute("y").toInt() );
		idxpos_initialized = true;
	}
	else if ( e.tagName() == "ArtWinGeom" && !artgeom_initialized ) {
		defArtX = e.attribute("x").toInt();
		defArtY = e.attribute("y").toInt();
		defArtWidth = e.attribute("w").toInt();
		defArtHeight = e.attribute("h").toInt();
		artgeom_initialized = true;
	}
  	n = n.nextSibling();
  }
  f.close();
}

void IdxView::saveConfig() {
  QPoint p = pos();
  if ( !p.x() && !p.y() ) p = mapToGlobal(p);

  QFile f( "config.xml" );
  QDomDocument doc( "pingvo" );
  QDomElement docEl, e;

  docEl = doc.createElement( "pingvo" );
  doc.appendChild( docEl );

  e = doc.createElement( "IdxWinPos" );
  e.setAttribute( "x", p.x()/* - (geometry().x() - x())*/);
  e.setAttribute( "y", p.y()/* - (geometry().y() - y())*/);
  docEl.appendChild( e );

  e = doc.createElement( "ArtWinGeom" );
  e.setAttribute( "x", defArtX );
  e.setAttribute( "y", defArtY );
  e.setAttribute( "w", defArtWidth );
  e.setAttribute( "h", defArtHeight );
  docEl.appendChild( e );

  if ( !f.open( IO_WriteOnly | IO_Truncate ) ) {
  	puts( "pingvo.cc config.xml open for writing failed" );
	return;
  }
  QTextStream stream( &f );
  doc.save( stream, 0 );
  f.close();

  configChanged = false;
}

void IdxView::confSaveTimeout() {
  delete configSaveTimer;
  configSaveTimer = 0;
  saveConfig();
}

void IdxView::articleClosed( int x, int y, int w, int h ) {
  if ( artBrowsersNum == 1 ) {
	defArtX = x;
	defArtY = y;
	defArtWidth = w;
	defArtHeight = h;
	saveConfig();
  }
  artBrowsersNum--;
}

void IdxView::xRecordEventsAvail() {
  recorder_process_replies();
}

void IdxView::processAnchorClick( const QString& word, const QString& ) {
  QListBoxItem *it;

  it = lb->firstItem();
  while ( it ) {
  	if ( it->text() == word ) {
		showArticle( it );
		break;
	}
	it = it->next();
  }
}

class RawLabel: public QWidget {
public:
  RawLabel( char *s );
  void paintEvent( QPaintEvent *e );
  void save( QString fname );
protected:
  char *s;
  int width, height;
  void updateContent();
};

RawLabel::RawLabel( char *s0 ): s(s0) {
  updateContent();
  resize( width, height );
  show();
}

void RawLabel::paintEvent( QPaintEvent *e ) {
  updateContent();
}

void RawLabel::updateContent() {
  GC gc;
  Font fnt;
  Display *d = x11Display();
  char *xlfd = "-windows-lingvo oem-medium-r-normal-sans-20-100-120-120-p-80-koi8-r";
  int dir, ascent, descent;
  XCharStruct overall;

  fnt = XLoadFont( d, xlfd );
  gc = XCreateGC( d, winId(), 0, 0 );
  XSetFont( d, gc, fnt );
  XSetForeground( d, gc, foregroundColor().pixel() );
  XSetBackground( d, gc, backgroundColor().pixel() );
  XQueryTextExtents( d, fnt, s, strlen(s), &dir, &ascent, &descent, &overall );
  XDrawImageString( d, winId(), gc, 0, ascent, s, strlen(s) );
  XUnloadFont( d, fnt );
  width = overall.width;
  height = ascent + descent;
}

void RawLabel::save( QString fname ) {
  QPixmap pm( size() );
  bitBlt( &pm, 0, 0, this, 0, 0, width, height );
  pm.save( fname, "XPM" );
}

ArticleView::ArticleView() {
//  enableClipper( TRUE );
//  viewport()->setEraseColor( QColor( 0xff, 0xff, 0xff ) );
//  show();
}

void ArticleView::loadArticle( char *fname, int start, int stop ) {
#if 0
  FILE *f;
  int x, y, rc;
  char *text;
  QLabel *l;

  f = fopen( fname, "r" );
  if ( !f ) {
	perror( "fopen" );
	return;
  }
  fseek( f, start, SEEK_SET );
  while ( !feof(f) && (stop == -1 || ftell(f) < stop) ) {
	rc = fscanf( f, "%d %d {%a[^}]}\n", &x, &y, &text );
	if ( rc != 3 ) break;
	l = new QLabel( QString::fromLocal8Bit(text), viewport() );
	l->setEraseColor( QColor( 0xff, 0xff, 0xff ) );
	l->adjustSize();
	addChild( l, x, y );
  }
  fclose( f );

  QRect r = viewport()->childrenRect();
  resizeContents( r.width(), r.height() );
#endif
}

QString ArticleView::transXMLtoRichText( QDomElement e ) {
  QDomNode n;
  QString text;
  
  for ( n = e.firstChild(); !n.isNull(); n = n.nextSibling() ) {
  	if ( n.isText() )
		text += n.toText().data();
	else if ( n.isElement() ) {
		e = n.toElement();
#if 0
		if ( e.tagName() == "m1" )
			text += "<p>&nbsp;" + transXMLtoRichText( e ) +
				"</p>";
		else if ( e.tagName() == "m2" )
			text += "<p>&nbsp;&nbsp;" + transXMLtoRichText( e ) +
				"</p>";
#else
		if ( e.tagName() == "m1" )
			text += "<tr><td>&nbsp;" + transXMLtoRichText( e ) +
				"</td></tr>";
		else if ( e.tagName() == "m2" )
			text += "<tr><td>&nbsp;&nbsp;" + transXMLtoRichText( e ) +
				"</td></tr>";
#endif
		else if ( e.tagName() == "com" )
			text += "<i><font color=\"#007F00\">" +
				transXMLtoRichText( e ) + "</font></i>\n";
		else if ( e.tagName() == "ex" )
			text += "<font color=\"#7F7F7F\">" +
				transXMLtoRichText( e ) + "</font>\n";
		else if ( e.tagName() == "trn" )
			text += transXMLtoRichText( e );
		else if ( e.tagName() == "b" )
			text += "<b>" + transXMLtoRichText( e ) +
				"</b>";
		else if ( e.tagName() == "ref" ) {
			QString ref = transXMLtoRichText( e );
			text += "<a name=\"" + ref +"\" href=\"#\">" +
				ref + "</a>";
		}
		else if ( e.tagName() == "t" ) {
			text += "<tr><td>[" + transXMLtoRichText( e ) + "]</td></tr>";
/*
			char p[80], *p1, *s, *s1;

			p1 = p;
			*p1++ = '[';
			s = (char*) t.latin1();
			for (;;) {
				*p1 = strtol( s, &s1, 10 );
				if ( s == s1 ) break;
				s = s1;
				if ( *s == ' ' ) s++;
			}
			*p1++ = ']';
			*p1 = 0;
			RawLabel rl( p  );
			rl.save( "test.xpm" );
			text += "<img src=\"test.xpm\">";
*/
		}
		else
			printf( "element with tag: %s\n", e.tagName().latin1() );
	}
  }

  return text;
}

void ArticleView::loadArticleXML( QDomElement e ) {
  QString text =
  	"<table border=0 cellspacing=0 cellpadding=0>"
	"<tr><td><h2>" + e.attribute("word") + "</h2></td></tr>" +
	transXMLtoRichText( e ) + "</table>";

  setText( text );
  show();
}

void ArticleView::closeEvent( QCloseEvent *e ) {
  QPoint p = pos();
  if ( !p.x() && !p.y() ) {
  	puts( "ArticleView doing mapToGlobal" );
  	p = mapToGlobal(p);
  }

  geometryOnClose( p.x(), p.y(), width(), height() );
  e->accept();
}

IdxView *idxview;

void magicReqHandler() {
  QClipboard *cb = qApp->clipboard();
  QString item = cb->text( QClipboard::Selection ).section( ' ', 0, 0 );

  idxview->le->setText( item.lower() );
  idxview->lookupByPrefix();
  if ( idxview->lb->currentText().lower() == item.lower() )
	idxview->showCurrentArticle();
}

int main( int argc, char **argv ) {
  QApplication qa( argc, argv );

  IdxView w;
  idxview = &w;
  qa.setMainWidget( &w );

  return qa.exec();
}
