#include <qscrollview.h>
#include <qsocketnotifier.h>
#include <qtimer.h>
#include <qdom.h>
#include "design/idxview.h"

class IdxViewItem: public QListBoxText {
public:
  int idx, art_start, art_end;
  QString tstr;

  IdxViewItem *prevSibling, *nextSibling;
  IdxViewItem *firstChild;

  IdxViewItem( int start, int end, char *text );
  virtual QString text() const;
};

class IdxView: public IdxViewUI {
  Q_OBJECT
public:
  IdxView( QWidget *parent = 0, const char *name = 0, WFlags f = 0 );
  ~IdxView();
  void loadDict( QString prefix );
  void loadDictXML( QString prefix );
protected:
  QString dname, idxname, wfname;
  int artBrowsersNum, defArtX, defArtY, defArtWidth, defArtHeight;
  bool xmlMode, configChanged, configured;
  QTimer *configSaveTimer;
  QSocketNotifier *ev_sn;

  IdxViewItem *rootNode;
  IdxViewItem *lookupNode( IdxViewItem *from, QString prefix, int pfxlen );
  IdxViewItem *lookupNode( QString prefix );
  void insertNode( IdxViewItem *node );
  void dumpChilds( IdxViewItem *node, int indent );
  void dumpTree();

  void loadConfig();
  void saveConfig();
  QDomElement getArticleXML( IdxViewItem *it );
  virtual void keyPressEvent( QKeyEvent *e );
  virtual void moveEvent( QMoveEvent *e );
  virtual void closeEvent( QCloseEvent *e );
public slots:
  void lookupByPrefix();
  void loadEnRu();
  void loadRuEn();
  void loadDefXML();
  void loadCustom();
  void showArticle( QListBoxItem *lbItem );
  void showCurrentArticle();
  void xRecordEventsAvail();
  void confSaveTimeout();
  void articleClosed( int x, int y, int w, int h );
  void processAnchorClick( const QString& word, const QString& link );
};

class ArticleView: public QTextBrowser {
  Q_OBJECT
public:
  ArticleView();
  void loadArticle( char *fname, int start, int stop );
  void loadArticleXML( QDomElement e );
protected:
  QString transXMLtoRichText( QDomElement e );
  void closeEvent( QCloseEvent *e );
signals:
  void geometryOnClose( int x, int y, int w, int h );
};

